/*
 * Tapiocaui app
 * Copyright (C) 2005 INdT
 * @author Renato Araujo Oliveira Filho <renato.filho@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gtk/gtk.h>
#include <glib/ghash.h>
#include <glib/gstdio.h>

#include "tapiocaui_shell.h"
#include "tapiocaui_shell_priv.h"
#include "tapiocaui_shell_media.h"
#include "tapiocaui_shell_session.h"
#include "tapiocaui_shell_mainwin.h"
#include "tapiocaui_shell_presence.h"
#include "tapiocaui_shell_contact_list.h"
#include "tapiocaui_shell_communication.h"

// shell functions
static void        tapiocaui_shell_class_init                        (TapiocaUIShellClass *k);
static void        tapiocaui_shell_init                              (TapiocaUIShell *self);
static void        tapiocaui_shell_finalize                          (GObject *object);
static gboolean    tapiocaui_shell_call_buddy_find                   (gpointer key,
                                                                      gpointer value,
                                                                      gpointer data);

G_DEFINE_TYPE (TapiocaUIShell, tapiocaui_shell, G_TYPE_OBJECT)

static void
tapiocaui_shell_class_init (TapiocaUIShellClass *k)
{
    GObjectClass *object_class = G_OBJECT_CLASS (k);
    tapiocaui_shell_parent_class = g_type_class_peek_parent (k);
    object_class->finalize = tapiocaui_shell_finalize;
}

static void
tapiocaui_shell_init (TapiocaUIShell *self)
{
    char *dir_name;

    self->priv = g_new0 (TapiocaUIShellPrivate, 1);

    // create conf dir
    dir_name = g_build_filename (g_get_home_dir (),
                                 ".tapioca", NULL);
    if (!g_file_test (dir_name, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))
        if (g_mkdir (dir_name, 0750) != 0)
            g_warning ("Error creating dir $HOME/.tapioca");

    g_free (dir_name);
}

static void
tapiocaui_shell_finalize (GObject *object)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (object);
    g_free (self->priv);

    G_OBJECT_CLASS (tapiocaui_shell_parent_class)->finalize (object);
}

TapiocaUIShell *
tapiocaui_shell_new (void)
{
    TapiocaUIShell *self;
    self = g_object_new (TAPIOCAUI_TYPE_SHELL, NULL);
    return self;
}

gboolean
tapiocaui_shell_initialize (TapiocaUIShell *self)
{
    self->priv->db_account = tapiocaui_db_new ("accounts.xml");
    tapiocaui_db_load (self->priv->db_account);
    self->priv->win = GTK_WIDGET (tapiocaui_main_win_new (
                self->priv->db_account));
    self->priv->active_comm_list =  g_hash_table_new_full (
            g_str_hash, g_str_equal, g_free, NULL);

    g_signal_connect (G_OBJECT (self->priv->win),
                      "delete-event",
                      G_CALLBACK (tapiocaui_shell_mainwin_delete_cb),
                      NULL);
    g_signal_connect (G_OBJECT (self->priv->win),
                      "login",
                      G_CALLBACK (tapiocaui_shell_mainwin_login_cb),
                      self);
    g_signal_connect (G_OBJECT (self->priv->win),
                      "login-cancelled",
                      G_CALLBACK (tapiocaui_shell_mainwin_login_cancelled_cb),
                      self);
    g_signal_connect (G_OBJECT (self->priv->win),
                      "logout",
                      G_CALLBACK (tapiocaui_shell_mainwin_logout_cb),
                      self);
    g_signal_connect (G_OBJECT (self->priv->win),
                      "call",
                      G_CALLBACK (tapiocaui_shell_mainwin_call_cb),
                      self);
    g_signal_connect (G_OBJECT (self->priv->win),
                      "open-chat",
                      G_CALLBACK (tapiocaui_shell_mainwin_open_chat_cb),
                      self);

    // initialize tapioca
    if (!(self->priv->tcm = tapioca_connection_manager_new ())) {
        tapiocaui_debug ("Error creating connection manager object. Aborting!");
        exit (1);
    }

    return TRUE;
}

void
tapiocaui_shell_terminate (TapiocaUIShell *self)
{
    if (self->priv->tcm != NULL) {
        g_object_unref (self->priv->tcm);
        self->priv->tcm = NULL;
    }

    if (self->priv->tc != NULL) {
        g_object_unref (self->priv->tc);
        self->priv->tc = NULL;
    }

    g_hash_table_destroy (self->priv->active_comm_list);
    g_object_unref (self->priv->db_account);

    g_free (self->priv);
}

TapiocaUIShellActiveComm *
tapiocaui_shell_get_communication (TapiocaUIShell *self,
                                   const gchar *jid)
{
    TapiocaUIShellActiveComm *comm;
    TapiocaUIBuddyData *buddy;
    TapiocaUIBuddyPresence *buddy_presence;
    gchar *bare_jid;

    tapiocaui_debug ("searching for active communication for user %s on hash table %p",
                     jid, self->priv->active_comm_list);

    bare_jid = tapiocaui_get_bare_jid (jid);

    comm = g_hash_table_lookup (self->priv->active_comm_list, bare_jid);

    if (comm) {
        tapiocaui_debug ("communication found");
        g_free (bare_jid);
        return comm;
    }

    comm = g_new0 (TapiocaUIShellActiveComm, 1);
    comm->mutex = g_mutex_new ();
    comm->idle_source = -1;
    comm->last_notification_state = TAPIOCA_TEXT_SESSION_NOTIFICATION_STATE_NONE;

    tapiocaui_debug ("creating communication");
    buddy = tapiocaui_db_find_buddy_by_id (self->priv->db_buddy,
                                           TAPIOCAUI_DB_ENTRY_BUDDY,
                                           bare_jid);

    if (buddy && (g_hash_table_size (buddy->jids) == 0)) {
        g_free (bare_jid);
        return NULL;
    }

    if (buddy == NULL) {
        tapiocaui_debug ("unknown buddy calling");
        buddy = tapiocaui_db_entry_new (TAPIOCAUI_DB_ENTRY_BUDDY);
        buddy->name = g_strdup (bare_jid);
        // buddy->jids = g_list_append (buddy->jids, jid);
        buddy->id = g_strdup (bare_jid);
        buddy->group = g_strdup ("");
        buddy_presence = g_new0 (TapiocaUIBuddyPresence, 1);
        buddy_presence->status = TAPIOCA_PRESENCE_STATUS_AVAILABLE;
        buddy_presence->status_desc = NULL;
        g_hash_table_insert (buddy->jids, g_strdup (jid), buddy_presence);
    }
    tapiocaui_debug ("request chat window");
    comm->comm_win = tapiocaui_main_win_open_chat (TAPIOCAUI_MAIN_WIN (self->priv->win),
                                                   buddy);
    comm->buddy = buddy;
    tapiocaui_debug ("insert communication %p with user %s on hash table %p",
                     comm, bare_jid, self->priv->active_comm_list);
    g_hash_table_insert (self->priv->active_comm_list, g_strdup (bare_jid), comm);

    g_signal_connect (G_OBJECT (comm->comm_win),
                      "send-message",
                      G_CALLBACK (tapiocaui_shell_communication_send_message_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "composing",
                      G_CALLBACK (tapiocaui_shell_communication_composing_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-answer",
                      G_CALLBACK (tapiocaui_shell_communication_call_answer_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-reject",
                      G_CALLBACK (tapiocaui_shell_communication_call_reject_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-disconnect",
                      G_CALLBACK (tapiocaui_shell_communication_call_disconnect_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-make",
                      G_CALLBACK (tapiocaui_shell_communication_call_make_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "quit",
                      G_CALLBACK (tapiocaui_shell_communication_quit_cb),
                      self);

    g_free (bare_jid);

    return comm;
}

void
tapiocaui_shell_complete_login (TapiocaUIShell *self,
                                const gchar *user)
{
    gchar *db_name;

    db_name = g_strconcat (user, "_db.xml", NULL);
    self->priv->db_buddy = tapiocaui_db_new (db_name);
    tapiocaui_db_load (self->priv->db_buddy);

    /* Resquest the remote contact list */
    tapioca_contact_list_request_contacts (self->priv->contact_list);

    tapiocaui_main_win_load_buddy (TAPIOCAUI_MAIN_WIN (self->priv->win),
                                   self->priv->db_buddy,
                                   user);

    g_free (db_name);
}

void
tapiocaui_shell_call_buddy (TapiocaUIShell *self,
                            TapiocaUIShellActiveComm *active_comm)
{
    TapiocaAudioSession *session = NULL;
    gchar *local_cap = NULL;

    g_return_if_fail (self != NULL);
    g_return_if_fail (active_comm != NULL);

    if (active_comm->audio_session)
        return;

    session = (TapiocaAudioSession *) tapioca_connection_create_session (self->priv->tc, TAPIOCA_SESSION_TYPE_AUDIO);
    g_assert (session != NULL);

    /* Connect session signals */
    g_signal_connect (G_OBJECT (session),
                      "member-added",
                      G_CALLBACK (tapiocaui_shell_session_member_added_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "member-removed",
                      G_CALLBACK (tapiocaui_shell_session_member_removed_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "add-member-error",
                      G_CALLBACK (tapiocaui_shell_session_add_member_er_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "state-changed",
                      G_CALLBACK (tapiocaui_shell_session_state_changed_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "error",
                      G_CALLBACK (tapiocaui_shell_session_error_signal_cb),
                      self);

    active_comm->audio_session = session;

    setup_rtp_session (self, active_comm);
    setup_rtp_stream (self, active_comm);

    local_cap = create_sdp_from_codecs (farsight_stream_get_local_codecs (active_comm->stream));

    tapiocaui_debug ("set local capabilities %s", local_cap);

    if (!(tapioca_audio_session_set_local_capabilities (
                    active_comm->audio_session,
                    (const gchar *) local_cap))) {
        tapiocaui_debug ("error on set local capabilities");
        g_object_unref (active_comm->audio_session);
        active_comm->audio_session = NULL;
        g_object_unref (active_comm->session);
        active_comm->session = NULL;
        g_object_unref (active_comm->stream);
        active_comm->stream = NULL;
        g_free (local_cap);
        return;
    }

    g_free (local_cap);

    /* Connect audio session signals */
    g_signal_connect (G_OBJECT (active_comm->audio_session),
                      "remote-capabilities-updated",
                      G_CALLBACK (remote_capabilities_updated_cb),
                      active_comm);
    g_signal_connect (G_OBJECT (active_comm->audio_session),
                      "remote-candidate-added",
                      G_CALLBACK (remote_candidate_added_cb),
                      active_comm);

    if (!g_hash_table_find (active_comm->buddy->jids,
                       tapiocaui_shell_call_buddy_find,
                       (gpointer) active_comm->audio_session)) {
        tapiocaui_debug ("buddy cannot receive calls");
        g_object_unref (active_comm->audio_session);
        active_comm->audio_session = NULL;
        g_object_unref (active_comm->session);
        active_comm->session = NULL;
        g_object_unref (active_comm->stream);
        active_comm->stream = NULL;
        return;
    }

    tapiocaui_communication_set_status (active_comm->comm_win,
            TAPIOCAUI_CALL_RINGING);
}

void
tapiocaui_shell_open_call (TapiocaUIShell *self,
                           TapiocaUIShellActiveComm *comm)
{
    tapiocaui_debug ("open call %p", comm->audio_session);

    self->priv->in_use = TRUE;

    if (self->priv->active_comm) {
        tapiocaui_debug ("hold active call");
        // tapioca_audio_session_hold (self->priv->comm->audio_session);
        tapiocaui_communication_set_status (self->priv->active_comm->comm_win, TAPIOCAUI_CALL_HOLD);
        self->priv->active_comm = NULL;
    }

    self->priv->active_comm = comm;

    tapiocaui_communication_set_status (comm->comm_win,
                                        TAPIOCAUI_CALL_ANSWERED);
    tapiocaui_communication_set_volume (comm->comm_win, 50);

    g_signal_connect (G_OBJECT (comm->comm_win),
                      "volume-change",
                      G_CALLBACK (tapiocaui_shell_communication_volume_change_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "input-sound-mute",
                      G_CALLBACK (tapiocaui_shell_communication_input_mute_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "input-sound-unmute",
                      G_CALLBACK (tapiocaui_shell_communication_input_unmute_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-hold",
                      G_CALLBACK (tapiocaui_shell_communication_call_hold_cb),
                      comm->audio_session);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-unhold",
                      G_CALLBACK (tapiocaui_shell_communication_call_unhold_cb),
                      comm->audio_session);
}

void
tapiocaui_shell_close_call (TapiocaUIShell *self,
                            TapiocaUIShellActiveComm *comm)
{
    tapiocaui_communication_set_status (comm->comm_win, TAPIOCAUI_CALL_IDLE);

    g_signal_handlers_disconnect_by_func (G_OBJECT (comm->comm_win),
                                          G_CALLBACK (tapiocaui_shell_communication_volume_change_cb),
                                          self);
    g_signal_handlers_disconnect_by_func (G_OBJECT (comm->comm_win),
                                          G_CALLBACK (tapiocaui_shell_communication_input_mute_cb),
                                          comm->audio_session);
    g_signal_handlers_disconnect_by_func (G_OBJECT (comm->comm_win),
                                          G_CALLBACK (tapiocaui_shell_communication_input_unmute_cb),
                                          comm->audio_session);
    g_signal_handlers_disconnect_by_func (G_OBJECT (comm->comm_win),
                                          G_CALLBACK (tapiocaui_shell_communication_call_hold_cb),
                                          comm->audio_session);
    g_signal_handlers_disconnect_by_func (G_OBJECT (comm->comm_win),
                                          G_CALLBACK (tapiocaui_shell_communication_call_unhold_cb),
                                          comm->audio_session);

    if (comm->stream)
        close_rtp_stream (self, comm);

    if (comm->session)
        close_rtp_session (self, comm);

    if (comm->audio_session) {
        /* Check if session is closed before trying to close */
        if (tapioca_session_get_state (TAPIOCA_SESSION (comm->audio_session)) !=
            TAPIOCA_SESSION_STATE_CLOSED) {
            tapiocaui_debug ("closing audio session ...");
            tapioca_session_close (TAPIOCA_SESSION (comm->audio_session));
        }
        else
            comm->audio_session = NULL;
    }

    comm->remote_codecs_set = FALSE;

    self->priv->active_comm = NULL;
    self->priv->in_use = FALSE;
}

static gboolean
tapiocaui_shell_call_buddy_find (gpointer key,
                                 gpointer value,
                                 gpointer data)
{
    TapiocaAudioSession *audio = TAPIOCA_AUDIO_SESSION (data);
    gchar **members = g_new0 (gchar *, 2);

    members[0] = g_strdup (key);

    if (!(tapioca_session_add_members (TAPIOCA_SESSION (audio),
                                       (const gchar **) members))) {
        tapiocaui_debug ("error on add_member %s", members[0]);
        g_strfreev (members);
        return FALSE;
     }

    if (!tapioca_session_initiate (TAPIOCA_SESSION (audio))) {
        tapiocaui_debug ("error on initiate audio session with member %s", members[0]);
        tapioca_session_remove_members (TAPIOCA_SESSION (audio),
                                        (const gchar **) members);
        g_strfreev (members);
        return FALSE;
    }

    g_strfreev (members);

    return TRUE;
}
