/*
 * Tapiocaui app
 * Copyright (C) 2005 INdT
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

#include "tapiocaui_shell_media.h"

/* NAT traversal stuff */
static void        stream_error                                  (FarsightStream *stream,
                                                                  FarsightStreamError error,
                                                                  const gchar *debug,
                                                                  gpointer user_data);
static void        session_error                                 (FarsightSession *stream,
                                                                  FarsightSessionError error,
                                                                  const gchar *debug,
                                                                  gpointer user_data);
static void        new_active_candidate_pair                     (FarsightStream *stream,
                                                                  gchar *native_candidate,
                                                                  gchar *remote_candidate,
                                                                  gpointer user_data);
static void        codec_changed                                 (FarsightStream *stream,
                                                                  gint codec_id,
                                                                  gpointer user_data);
static void        new_native_candidate                          (FarsightStream *stream,
                                                                  const gchar *candidate_id,
                                                                  gpointer user_data);
static void        native_candidates_prepared                    (FarsightStream *stream,
                                                                  gpointer user_data);
static void        state_changed                                 (FarsightStream *stream,
                                                                  FarsightStreamState state,
                                                                  FarsightStreamDirection dir,
                                                                  gpointer user_data);

void
setup_rtp_session (TapiocaUIShell *self,
                   TapiocaUIShellActiveComm *comm)
{
    comm->session = farsight_session_factory_make ("rtp");
    if (!comm->session) {
        g_error ("RTP plugin not found");
        exit (1);
    }

    tapiocaui_debug ("protocol details:\n name: %s\n description: %s\n author: %s\n",
            farsight_plugin_get_name (comm->session->plugin),
            farsight_plugin_get_description (comm->session->plugin),
            farsight_plugin_get_author (comm->session->plugin));
    g_signal_connect (G_OBJECT (comm->session), "error",
            G_CALLBACK (session_error), NULL);
}

void
setup_rtp_stream (TapiocaUIShell *self,
                  TapiocaUIShellActiveComm *comm)
{
    const GList *possible_codecs, *lp;
    FarsightCodec *codec;
    GstElement *srcbin;
    GstElement *alsasrc;
    GstPad *srcpad;
    /* GstElement *sinkbin; */
    GstElement *alsasink;
    /* GstPad *sinkpad; */

    comm->stream = farsight_session_create_stream (comm->session,
            FARSIGHT_MEDIA_TYPE_AUDIO, FARSIGHT_STREAM_DIRECTION_BOTH);
    tapiocaui_debug ("setting up stream %p with self %p and comm %p",
            comm->stream, self, comm);
    g_signal_connect (G_OBJECT (comm->stream), "error",
                      G_CALLBACK (stream_error), self);
    g_signal_connect (G_OBJECT (comm->stream), "new-active-candidate-pair",
                      G_CALLBACK (new_active_candidate_pair), comm);
    g_signal_connect (G_OBJECT (comm->stream), "codec-changed",
                      G_CALLBACK (codec_changed), comm);
    g_signal_connect (G_OBJECT (comm->stream), "native-candidates-prepared",
                      G_CALLBACK (native_candidates_prepared), comm);
    g_signal_connect (G_OBJECT (comm->stream), "new-native-candidate",
                      G_CALLBACK (new_native_candidate), comm);
    g_signal_connect (G_OBJECT (comm->stream), "state-changed",
                      G_CALLBACK (state_changed), self);

    possible_codecs = farsight_stream_get_local_codecs (comm->stream);
    for (lp = possible_codecs; lp; lp = g_list_next (lp)) {
        codec = (FarsightCodec*) lp->data;
        tapiocaui_debug ("codec: %d: %s/%d found",
                codec->id, codec->encoding_name,
                codec->clock_rate);
    }

    /* create src bin */
    srcbin = gst_bin_new ("src");
    alsasrc = gst_element_factory_make ("alsasrc", "alsasrc");
    g_object_set (G_OBJECT (alsasrc), "blocksize", 320, NULL);
    g_object_set (G_OBJECT (alsasrc), "latency-time", G_GINT64_CONSTANT (20000), NULL);
    /* default output volume to 100% */
    comm->srcvolume = gst_element_factory_make ("volume", "srcvolume");
    g_object_set (G_OBJECT (comm->srcvolume), "volume", 4.0, NULL);
    gst_bin_add_many (GST_BIN (srcbin), alsasrc, comm->srcvolume, NULL);
    gst_element_link (alsasrc, comm->srcvolume);
    srcpad = gst_ghost_pad_new ("src",
            gst_element_get_pad (comm->srcvolume, "src"));
    gst_element_add_pad (srcbin, srcpad);

    /* create sink bin */
    /* sinkbin = gst_bin_new ("sink"); */
    /* default output volume to 50% */
    /*
    comm->sinkvolume = gst_element_factory_make ("volume", "sinkvolume");
    g_object_set (G_OBJECT (comm->sinkvolume), "volume", 2.0, NULL);
    */
    alsasink = gst_element_factory_make ("alsasink", "alsasink");
    g_object_set (G_OBJECT (alsasink), "sync", FALSE, NULL);
    g_object_set (G_OBJECT (alsasink), "latency-time", G_GINT64_CONSTANT (20000), NULL);
    g_object_set (G_OBJECT (alsasink), "buffer-time", G_GINT64_CONSTANT (80000), NULL);
    /*
    gst_bin_add_many (GST_BIN (sinkbin), comm->sinkvolume, alsasink, NULL);
    gst_element_link (comm->sinkvolume, alsasink);
    sinkpad = gst_ghost_pad_new ("sink",
            gst_element_get_pad (comm->sinkvolume, "sink"));
    gst_element_add_pad (sinkbin, sinkpad);
    */

    farsight_stream_set_source (comm->stream, srcbin);
    /* farsight_stream_set_sink (comm->stream, sinkbin); */
    farsight_stream_set_sink (comm->stream, alsasink);
}

void
close_rtp_session (TapiocaUIShell *self,
                   TapiocaUIShellActiveComm *comm)
{
    FarsightSession *session = comm->session;

    tapiocaui_debug ("stoping media session!");
    g_signal_handlers_disconnect_by_func (G_OBJECT (session), G_CALLBACK (session_error), NULL);
    farsight_session_destroy (session);
    comm->session = NULL;
}

void
close_rtp_stream (TapiocaUIShell *self,
                  TapiocaUIShellActiveComm *comm)
{
    FarsightStream *stream = comm->stream;

    tapiocaui_debug ("stoping media stream!");
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (stream_error), self);
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (new_active_candidate_pair), comm);
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (codec_changed), comm);
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (native_candidates_prepared), comm);
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (new_native_candidate), comm);
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (state_changed), self);

    if (farsight_stream_get_state != FARSIGHT_STREAM_STATE_STOPPED)
        farsight_stream_stop (stream);
    g_object_unref (stream);
    comm->stream = NULL;
}

void
remote_candidate_added_cb (TapiocaAudioSession *audio_session,
                           const gchar *name,
                           const gchar *protocol,
                           const gchar *ip,
                           guint port,
                           gdouble preference,
                           const gchar *username,
                           const gchar *password,
                           const gchar *type,
                           guint generation,
                           gpointer user_data)
{
    TapiocaUIShellActiveComm *comm = (TapiocaUIShellActiveComm *) user_data;
    GList *candidate = NULL;
    FarsightTransportInfo *info;

    info = g_new0 (FarsightTransportInfo, 1);
    info->candidate_id = g_strdup_printf ("L%d", ++comm->candidate_id);
    info->component = 1;
    info->ip = g_strdup (ip);
    info->port = port;
    info->proto = (g_strcasecmp (protocol, "tcp") == 0) ? FARSIGHT_NETWORK_PROTOCOL_TCP : FARSIGHT_NETWORK_PROTOCOL_UDP;
    info->proto_subtype = g_strdup ("RTP");
    info->proto_profile = g_strdup ("AVP");
    info->preference = preference;
    if (!g_strcasecmp (type, "local"))
        info->type = FARSIGHT_CANDIDATE_TYPE_LOCAL;
    else if (!g_strcasecmp (type, "stun"))
        info->type = FARSIGHT_CANDIDATE_TYPE_DERIVED;
    else if (!g_strcasecmp (type, "relay"))
        info->type = FARSIGHT_CANDIDATE_TYPE_RELAY;
    info->username = g_strdup (username);
    info->password = g_strdup (password);

    tapiocaui_debug ("remote transport candidate: %s %d %s (%d) %s %s:%d %s %s, pref %f",
                info->candidate_id,
                info->component,
                protocol,
                info->proto,
                info->proto_subtype,
                info->ip,
                info->port,
                info->username,
                info->password,
                (double) info->preference);

    candidate = g_list_append (candidate, info);
    farsight_stream_add_remote_candidate (comm->stream, candidate);

    farsight_transport_list_destroy (candidate);
}

static void
session_error (FarsightSession *stream,
               FarsightSessionError error,
               const gchar *debug,
               gpointer user_data)
{
    tapiocaui_debug ("session error: session=%p error=%s", stream, debug);
}

static void
stream_error (FarsightStream *stream,
              FarsightStreamError error,
              const gchar *debug,
              gpointer user_data)
{
    tapiocaui_debug ("stream error: stream=%p error=%s", stream, debug);
}

static void
new_active_candidate_pair (FarsightStream *stream,
                           gchar *native_candidate,
                           gchar *remote_candidate,
                           gpointer user_data)
{
    TapiocaUIShellActiveComm *comm = (TapiocaUIShellActiveComm *) user_data;

    tapiocaui_debug ("new native candidate pair: stream=%p", stream);

    g_mutex_lock (comm->mutex);

    if (comm->remote_codecs_set) {
        tapiocaui_debug ("starting media stream ...");
        farsight_stream_start (stream);
    }

    g_mutex_unlock (comm->mutex);
}

static void
codec_changed (FarsightStream *stream,
               gint codec_id,
               gpointer user_data)
{
    tapiocaui_debug ("codec changed: codec_id=%d, stream=%p", codec_id, stream);
}

static void
new_native_candidate (FarsightStream *stream,
                      const gchar *candidate_id,
                      gpointer user_data)
{
    TapiocaUIShellActiveComm *comm = (TapiocaUIShellActiveComm *) user_data;
    GList *native_candidate = NULL, *lc;
    FarsightTransportInfo *info;
    gchar *proto, *type = NULL;

    tapiocaui_debug ("new native candidate found");

    native_candidate = farsight_stream_get_native_candidate (stream, candidate_id);

    for (lc = native_candidate; lc; lc = g_list_next (lc)) {
        info = (FarsightTransportInfo *) lc->data;
        proto = (info->proto == FARSIGHT_NETWORK_PROTOCOL_TCP) ? "tcp" : "udp";
        if (info->type == FARSIGHT_CANDIDATE_TYPE_LOCAL)
            type = "local";
        else if (info->type == FARSIGHT_CANDIDATE_TYPE_DERIVED)
            type = "stun";
        else if (info->type == FARSIGHT_CANDIDATE_TYPE_RELAY)
            type = "relay";

        tapiocaui_debug ("local transport candidate: %s %d %s %s %s:%d %s %s, pref %f",
                info->candidate_id,
                info->component,
                proto,
                info->proto_subtype, info->ip,
                info->port,
                info->username, info->password,
                (double) info->preference);
        tapioca_audio_session_add_local_candidate (
                comm->audio_session,
                "rtp",
                // info->candidate_id,
                proto,
                info->ip,
                info->port,
                info->preference,
                info->username,
                info->password,
                type,
                0);
    }

    farsight_transport_list_destroy (native_candidate);
}

static void
native_candidates_prepared (FarsightStream *stream,
                            gpointer user_data)
{
    const GList *transport_candidates, *lp;
    FarsightTransportInfo *info;

    tapiocaui_debug ("candidates preparation complete: stream=%p", stream);

    transport_candidates = farsight_stream_get_native_candidate_list (stream);
    for (lp = transport_candidates; lp; lp = g_list_next (lp))
    {
        info = (FarsightTransportInfo*)lp->data;
        tapiocaui_debug ("local transport candidate: %s %d %s %s %s:%d, pref %f",
                info->candidate_id, info->component,
                (info->proto == FARSIGHT_NETWORK_PROTOCOL_TCP) ? "TCP" : "UDP",
                info->proto_subtype, info->ip, info->port,
                (double) info->preference);
    }
}

static void
state_changed (FarsightStream *stream,
               FarsightStreamState state,
               FarsightStreamDirection dir,
               gpointer user_data)
{
    TapiocaUIShell *self = (TapiocaUIShell *) user_data;

    switch (state) {
        case FARSIGHT_STREAM_STATE_STOPPED:
            tapiocaui_debug ("stream stopped: stream=%p", stream);
            tapiocaui_shell_close_call (self, self->priv->active_comm);
            break;
        case FARSIGHT_STREAM_STATE_PLAYING:
            tapiocaui_debug ("stream playing: stream=%p", stream);
            break;
        default:
            break;

    }
}

GList *
create_codecs_from_sdp (const gchar *sdp_)
{
    GList *codecs = NULL;
    FarsightCodec *codec;
    gchar *sdp;
    guint pt;
    gint rate, channels;
    gchar *encoding_name, *offset;

    // TODO check if the string is correctly formed

    // a=rtpmap:PT ENCODING_NAME/RATE[/CHANNELS]\r\n";
    sdp = strstr (sdp_, "a=rtpmap:");
    while (sdp) {
        sdp = strchr (sdp, ':') + 1;    // sdp = "pt ENCODING_NAME/RATE[/CHANNELS]
        pt = strtoul (sdp, &sdp, 10);   // sdp = " ENCODING_NAME/RATE[/CHANNELS]
        ++sdp;                          // sdp = "ENCODING_NAME/RATE[/CHANNELS]
        offset = strchr (sdp, '/');
        encoding_name = g_strndup (sdp, offset - sdp);
        sdp = offset + 1;              // sdp = "RATE[/CHANNELS]
        rate = strtoul (sdp, &sdp, 10); // sdp = " [/CHANNELS]
        if (*sdp == '/') {
            ++sdp;                      // sdp = CHANNELS
            channels = strtoul (sdp, NULL, 10);
        }
        else
            channels = -1;

        if (channels != -1)
            g_debug ("adding codec %d %s/%d/%d", pt, encoding_name, rate, channels);
        else
            g_debug ("adding codec %d %s/%d", pt, encoding_name, rate);

        codec = g_new0 (FarsightCodec, 1);
        farsight_codec_init (codec, pt, encoding_name, FARSIGHT_MEDIA_TYPE_AUDIO, rate);
        codecs = g_list_append (codecs, codec);

        g_free (encoding_name);

        sdp = strstr (sdp, "a=rtpmap:");
    }

    return codecs;
}

gchar *
create_sdp_from_codecs (const GList *codecs)
{
    const GList *lc;
    FarsightCodec *codec;
    GString *sdp = g_string_new ("v=0\n"
            "o=- 123456 654321 IN IP4 127.0.0.1\n"
			"s=-\n"
			"c=IN IP4 127.0.0.1\n"
			"t=0 0\n");
    GString *media = g_string_new ("m=audio 7080 RTP/AVP ");
    gsize pos = sdp->len;

    for (lc = codecs; lc; lc = g_list_next (lc)) {
        codec = (FarsightCodec *) lc->data;
        if (codec->media_type != FARSIGHT_MEDIA_TYPE_AUDIO)
            continue;

        g_string_append_printf (media, "%d ", codec->id);

        g_string_append_printf (sdp, "a=rtpmap:%d %s/%d\n",
                codec->id, codec->encoding_name, codec->clock_rate);
    }

    g_string_append_printf (media, "\n");

    g_string_insert_len (sdp, pos, media->str, media->len);
    
    g_string_free (media, TRUE);

    return g_string_free (sdp, FALSE);
}

/* here the media capabilitie will be processed */
void
remote_capabilities_updated_cb (TapiocaAudioSession *session,
                                const gchar *remote_cap,
                                gpointer user_data)
{
    TapiocaUIShellActiveComm *comm = (TapiocaUIShellActiveComm *) user_data;
    GList *codecs;

    tapiocaui_debug ("remote capabilities updated to %s", remote_cap);

    g_mutex_lock (comm->mutex);

    /* TODO check if its needed to stop and start the
     *      pipeline if it is already running.
     */
    codecs = create_codecs_from_sdp (remote_cap);
    farsight_stream_set_remote_codecs (comm->stream, codecs);
    farsight_codec_list_destroy (codecs);

    comm->remote_codecs_set = TRUE;

    if (farsight_stream_get_state (comm->stream) == FARSIGHT_STREAM_STATE_CONNECTED) {
        tapiocaui_debug ("starting media stream ...");
        farsight_stream_start (comm->stream);
    }

    g_mutex_unlock (comm->mutex);
}
